// @ myth - client - question.h
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk

// @@ begin

#ifndef QUESTION_HEADER
#define QUESTION_HEADER

// @@ include

#include "common/command.h"
#include "client/output.h"
#include "common/ipc.h"
#include "client/help.h"

struct PAGE;
struct MACRO;

// Some forward refrences...

struct COMMAND;
struct QUESTION_SYMBOL;
struct QUESTION_LETTER;

// @@ QUESTION

// @@@ class QUESTION

class QUESTION
{
protected:
    bool is_done;		// Is question answered?
    prompt_type prompt;		// What primary prompt to use?
public:
    virtual QUESTION* current()	// No subquestions per default
    { return NULL; }
    bool done()			// Is this subquestion answered?
    { return is_done; }
    // Read the answer from keyboard
    virtual const char* read(MACRO*, long&);
    virtual void answer(CHANNEL*); // Send answer to channel
    // Update screen showing this question
    virtual void update(OUTPUT*, long, const char*);
    virtual void help(OUTPUT*);	// Tell player how to answer this
    // Handle local commands
    virtual bool check(OUTPUT*, long);
    QUESTION(prompt_type prompt);
    virtual ~QUESTION() { }
};

// @@@ class QUESTION_OBJECT

class QUESTION_OBJECT : public QUESTION
{
    prompt_type prompt;
    inventory_type bunch;
    QUESTION_SYMBOL* symbol;
    QUESTION_LETTER* letter;
public:
    QUESTION* current();
    void answer(CHANNEL* channel);
    QUESTION_OBJECT(prompt_type prompt, inventory_type bunch);
    ~QUESTION_OBJECT();
};    

// @@@ class QUESTION_POSITION

class QUESTION_POSITION : public QUESTION
{
    coor position;
public:
    coor where()		// Used by recenter command
    { return position; }
    const char* read(MACRO* macro, long&);
    void answer(CHANNEL* channel)
    { channel->PutCoor(position); }
    void update(OUTPUT* output, long argument, const char* error);
    void help(OUTPUT* output)
    { output->cover(::help.select(REQUEST_POSITION)); }
    QUESTION_POSITION(prompt_type prompt) : (prompt)
    { position = Origo; }
};

// @@@ class QUESTION_STRING

class QUESTION_STRING : public QUESTION
{
protected:
    char string[MessageSize];
public:
    const char* read(MACRO* macro, long&);
    void answer(CHANNEL* channel)
    { channel->PutString(string); }
    void update(OUTPUT* output, long argument, const char* error);
    void help(OUTPUT* output)
    { output->cover(::help.select(REQUEST_STRING)); }
    QUESTION_STRING(prompt_type prompt) : (prompt)
    { string[0] = 0; }
};

// @@@ class QUESTION_DIRECTION

class QUESTION_DIRECTION : public QUESTION
{
    direction dir;
public:
    const char* read(MACRO* macro, long&);
    void answer(CHANNEL* channel)
    { channel->PutChar(dir); }
    void update(OUTPUT* output, long argument, const char* error);
    void help(OUTPUT* output)
    { output->cover(::help.select(REQUEST_DIRECTION)); }
    QUESTION_DIRECTION(prompt_type prompt) : (prompt)
    { dir = DIRECTION_ERROR; }
};

// @@@ class QUESTION_SYMBOL

class QUESTION_SYMBOL : public  QUESTION
{
    inventory_type bunch;
    symbol_type symbol;
public:
    symbol_type what()		// Used by inventory command
    { return symbol; }
    const char* read(MACRO* macro, long&);
    void answer(CHANNEL* channel)
    { channel->PutChar(symbol); }
    void update(OUTPUT* output, long argument, const char* error);
    void help(OUTPUT* output)
    { output->cover(::help.select(REQUEST_SYMBOL)); }
    QUESTION_SYMBOL(prompt_type prompt, inventory_type bunch) : (prompt)
    { QUESTION_SYMBOL::bunch = bunch; symbol = SYMBOL_ERROR; }
};

// @@@ class QUESTION_CONFIRMATION

class QUESTION_CONFIRMATION : public QUESTION
{
    confirmation conf;
public:
    const char* read(MACRO* macro, long&);
    void answer(CHANNEL* channel);
    void update(OUTPUT* output, long argument, const char* error);
    void help(OUTPUT* output)
    { output->cover(::help.select(REQUEST_CONFIRMATION)); }
    QUESTION_CONFIRMATION(prompt_type prompt) : (prompt)
    { conf = CONFIRMATION_ERROR; }
};

// @@@ class QUESTION_ARGUMENT

class QUESTION_ARGUMENT : public QUESTION
{
public:
    const char* read(MACRO* macro, long& argument);
    void answer(CHANNEL* channel) { }
    void update(OUTPUT* output, long argument, const char* error);
    void help(OUTPUT* output)
    { output->cover(::help.select(REQUEST_ARGUMENT)); }
    QUESTION_ARGUMENT(prompt_type prompt) : (prompt) { }
};

// @@@ class QUESTION_LETTER

class QUESTION_LETTER : public QUESTION
{
    inventory_type bunch;
    symbol_type symbol;
    int letter;
public:
    int what()			// Tell QUESTION_OBJECT
    { return letter; }
    const char* read(MACRO* macro, long&);
    void answer(CHANNEL* channel)
    { channel->PutChar(letter); }
    void update(OUTPUT* output, long argument, const char* error);
    void help(OUTPUT* output);
    QUESTION_LETTER(prompt_type prompt,
		    inventory_type bunch, symbol_type symbol);
};

// @@@ class QUESTION_NAME

class QUESTION_NAME : public QUESTION
{
    char string[MessageSize];
public:
    const char* read(MACRO* macro, long&);
    void answer(CHANNEL* channel)
    { channel->PutString(string); }
    void update(OUTPUT* output, long argument, const char* error);
    void help(OUTPUT* output)
    { output->cover(::help.select(REQUEST_NAME)); }
    QUESTION_NAME(prompt_type prompt);
};

// @@@ class QUESTION_COMMAND

class QUESTION_COMMAND : public QUESTION
{
protected:
    COMMAND* command;
public:
    QUESTION* current();
    const char* read(MACRO* macro, long&);
    void answer(CHANNEL* channel);
    void update(OUTPUT* output, long argument, const char* error);
    void help(OUTPUT* output)
    { output->cover(::help.select(REQUEST_COMMAND)); }
    bool check(OUTPUT* output, long);
    QUESTION_COMMAND(prompt_type prompt) : (prompt)
    { command = NULL; }
    ~QUESTION_COMMAND();
};

// @@@ class QUESTION_EXTENDED

class QUESTION_EXTENDED : public QUESTION_STRING
{
public:
    bool check(OUTPUT*, long);
    QUESTION_EXTENDED(prompt_type prompt) : (prompt) { }
};    

// @@@ class QUESTION_GAME_OVER

class QUESTION_GAME_OVER : public QUESTION_COMMAND
{
public:
    const char* read(MACRO* macro, long&);
    QUESTION_GAME_OVER() : (PROMPT_COPYRIGHT) { }
};

// @@ COMMAND

// @@@ class COMMAND

class COMMAND
{
public:
    bool done()
    { return current() == NULL; }
    virtual QUESTION* current()
    { return NULL; /* ..., for commands without arguments */ }
    virtual bool check(OUTPUT*, long);
    virtual void answer(CHANNEL*)
    { /* No action for local commands */ }
    COMMAND()
    { /* CC complained about "destructor without constructer" */ }
    virtual ~COMMAND()
    { /* Do Nothing */ }
};

// @@@ class COMMAND_EXTERN
//
// Common code for extern commands (ie commands send to the server process)

class COMMAND_EXTERN : public COMMAND
{
protected:
    QUESTION* arg[MaxCommandLength]; // Arguments to the server command
    virtual from_client type()  // Each command has it's own type
    { return CLIENT_ERROR; /* Should not be called! */ }
public:
    QUESTION* current();	// Use arg to find current command
    bool check(OUTPUT* output, long argument);
    void answer(CHANNEL* channel); // Send type and arg to server
    COMMAND_EXTERN();	   	// NULL'ify arg array
    ~COMMAND_EXTERN();	   	// Delete arg array
};

// @@@ class COMMAND_MOVE

class COMMAND_MOVE : public COMMAND_EXTERN
{
public:
    from_client type()
    { return CLIENT_MOVE; }
    COMMAND_MOVE()
    { arg[0] = new QUESTION_DIRECTION(PROMPT_MOVE); }
};    

// @@@ class COMMAND_RUN

class COMMAND_RUN : public COMMAND_EXTERN
{
public:
    from_client type()
    { return CLIENT_RUN; }
    COMMAND_RUN()
    { arg[0] = new QUESTION_DIRECTION(PROMPT_RUN); }
};    

// @@@ class COMMAND_DROP

class COMMAND_DROP : public COMMAND_EXTERN
{
public:
    from_client type()
    { return CLIENT_DROP; }
    COMMAND_DROP()
    { arg[0] = new QUESTION_OBJECT(PROMPT_DROP, INVENTORY_OWN); }
};    

// @@@ class COMMAND_GET

class COMMAND_GET : public COMMAND_EXTERN
{
public:
    from_client type()
    { return CLIENT_GET; }
    bool check(OUTPUT*, long)
    { return FALSE; }
};    

// @@@ class COMMAND_APPLY

class COMMAND_APPLY : public COMMAND_EXTERN
{
public:
    from_client type()
    { return CLIENT_APPLY; }
    COMMAND_APPLY()
    { arg[0] = new QUESTION_OBJECT(PROMPT_APPLY, INVENTORY_OWN); }
};    

// @@@ class COMMAND_THROW

class COMMAND_THROW : public COMMAND_EXTERN
{
public:
    from_client type()
    { return CLIENT_THROW; }
    COMMAND_THROW()
    {
	arg[0] = new QUESTION_OBJECT(PROMPT_THROW, INVENTORY_OWN);
	arg[1] = new QUESTION_POSITION(PROMPT_THROW);
    }
};    

// @@@ class COMMAND_SLEEP

class COMMAND_SLEEP : public COMMAND_EXTERN
{
public:
    from_client type()
    { return CLIENT_SLEEP; }
    COMMAND_SLEEP()
    { arg[0] = new QUESTION_CONFIRMATION(PROMPT_SLEEP); }
};    

// @@@ class COMMAND_COMMAND

class COMMAND_COMMAND : public COMMAND_EXTERN
{
public:
    from_client type()
    { return CLIENT_COMMAND; }
    COMMAND_COMMAND()
    { arg[0] = new QUESTION_EXTENDED(PROMPT_EXTENDED); }
};    

// @@@ class COMMAND_LOCAL
//
// Common code for commands local to the client process.

class COMMAND_LOCAL : public COMMAND
{
public:
    // Actually run the local command
    virtual void execute(OUTPUT*, long);
    bool check(OUTPUT* output, long argument)
    { execute(output, argument); return TRUE; }
};

// @@@ class COMMAND_RECENTER

class COMMAND_RECENTER : public COMMAND_LOCAL
{
    QUESTION_POSITION* position;
public:
    QUESTION* current()
    { return position->done() ? NULL : position; }
    void execute(OUTPUT* output, long);
    COMMAND_RECENTER();
    ~COMMAND_RECENTER();
};    

// @@@ class COMMAND_INVENTORY

class COMMAND_INVENTORY : public COMMAND_LOCAL
{
    QUESTION_SYMBOL* symbol;
public:
    QUESTION* current()
    { return symbol->done() ? NULL : symbol; }
    void execute(OUTPUT* output, long argument)
    { output->inventory.show(INVENTORY_OWN, symbol->what(), argument); }
    COMMAND_INVENTORY();
    ~COMMAND_INVENTORY();
};    

// @@@ class COMMAND_STATUS

class COMMAND_STATUS : public COMMAND_LOCAL
{
public:
    void execute(OUTPUT* output, long argument)
    { output->status.show(argument); }
};    

// @@@ class COMMAND_NEWS

class COMMAND_NEWS : public COMMAND_LOCAL
{
public:
    void execute(OUTPUT* output, long);
};    

// @@@ class COMMAND_UP

class COMMAND_UP : public COMMAND_LOCAL
{
public:
    void execute(OUTPUT* output, long argument)
    { output->news.up(argument); }
};    

// @@@ class COMMAND_DOWN

class COMMAND_DOWN : public COMMAND_LOCAL
{
public:
    void execute(OUTPUT* output, long argument)
    { output->news.down(argument); }
};    

// @@ end

#endif QUESTION_HEADER

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
