// @ myth - server - party.C
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk

// @@ include

#include "server/party.h"
#include "common/error.h"
#include "common/parse.h"
#include "common/page.h"
#include "server/weapon.h"
#include "server/armor.h"
#include "server/food.h"
#include "server/player.h"
#include "common/format.h"

// @@ COMPACT

// @@@ class COMPACT

class COMPACT
{
    bool data[MaxChar];
public:
    void add(char letter)
    { data[letter] = TRUE; }
    char* string();
    void clear();
    COMPACT()
    { clear(); }
};

// @@@ char* COMPACT::string()

char* COMPACT::string()
{
    // Contain the string to be returned.
    static char message[MaxChar + 1]; // STATIC DATA! Be carefull!

    // State of the state machine
    enum
    {
	COMPACT_FIRST = -1,
	COMPACT_SECOND = -2,
	COMPACT_HYPHEN = -3,
	COMPACT_SKIP = -4
    }
         state = COMPACT_SKIP;

    int index = 0;		// Index in message string
    char last;			// Last letter found
    
    for(char current = FirstLetter;
	current != SYMBOL_BUNCH;
	current = NextLetter(current))
    {
	if(data[current])
	{
	    switch(state)
	    {
	    case COMPACT_SKIP:
		message[index++] = current;
		state = COMPACT_FIRST;
		break;
	    case COMPACT_FIRST:
		state = COMPACT_SECOND;
		last = current;
		break;
	    case COMPACT_SECOND:
		message[index++] = '-';
		state = COMPACT_HYPHEN;
		last = current;
		break;
	    case COMPACT_HYPHEN:
		last = current;
		break;
	    default:
		bug.error("COMPACT::string",
			  format("Unknown state %d", state));
		state = COMPACT_SKIP;
	    }
	}
	else
	{
	    if(state == COMPACT_SECOND || state ==  COMPACT_HYPHEN)
		message[index++] = last;

	    state = COMPACT_SKIP;
	}
    }
    if(state == COMPACT_SECOND || state ==  COMPACT_HYPHEN)
	message[index++] = last;

    message[index] = 0;
    
    return (index == 0) ? "*" : message;
}

// @@@ void COMPACT::clear()

void COMPACT::clear()
{
    for(int i = 0; i < MaxChar; i++)
	data[i] = FALSE;
}

// @@ OBJECT_LIST

// @@@ void OBJECT_LIST::erase()

void OBJECT_LIST::erase()
{
    delete current();	// Virtual delete function, I Hope.
    remove();
}

// @@@ void OBJECT_LIST::CreatePage(PAGE* page, PLAYER* player = NULL)
//
// Create a page describing objects in the list, adding entries which
// look equal together.
//
// See also INVENTORY::send.

void OBJECT_LIST::CreatePage(PAGE* page, PLAYER* player = NULL)
{
    sort();

    reset();
    while(remain() > 0)
    {
	STRING dummy = PackPage(current()->symbol(), page, player);
    }
}

// @@@ STRING OBJECT_LIST::PackPage(char symbol,
//
// PackPage is used by OBJECT_LIST::CreatePage and INVENTORY::send
// to make entries in the inventory which look similar, appear on the
// same line.
//
// It does this for a single symbol, call it repeatedly to get all
// objects in the list decribed.
//
// Inputs:
//
//   symbol: The symbol you want a list of
//   <list pointer>: Should point to the first symbol entry in the list
//
// Outputs:
//
//   <list pointer>: Point to the first entry after all symbol entries
//   FullPage: Text list of all symbol entries
//   <return>: String of letters used by symbol entries

STRING OBJECT_LIST::PackPage(char symbol, PAGE* FullPage, PLAYER* player)
{
    COMPACT AllLetters;
    PAGE page;			// All entries of this type
    OBJECT_LIST list;		// and their description

    // * Create a page containing one line for each entry in this list
    do
    {
	AllLetters.add(current()->letter());

	page.append(PrintObject(current(), PRINT_LOOK, player, 1L));
	list.append(current());
	next();
    }
    while(remain() > 0 && current()->symbol() == symbol);
    
    // * Transfer "unique" entries from page to FullPage
    do
    {
	// First unchecked entry
	list.reset();
	page.reset();

	STRING string = page.current();
	OBJECT* object = list.current();
	COMPACT compact;
	long count = 0L;

	// Check all other entries
	do
	{
	    if(page.current() == string)
	    {
		count += list.current()->count();
		compact.add(list.current()->letter());
		page.remove();
		list.remove();
	    }
	    else
	    {
		page.next();
		list.next();
	    }
	}
	while(page.remain() > 0);
	
	// Append entry to FullPage
	char look = IsKind(symbol) ? symbol : object->look();
	const char* text = ::print(object, PRINT_LOOK | PRINT_COUNT,
				   player, count);
	FullPage->append(format("%s%c %s", compact.string(), look, text));
    }
    while(page.length() > 0);

    // Return the letters found.
    return format("%s", (AllLetters.string()));
}

// @@@ void OBJECT_LIST::sort()

static comp compare(void* a, void* b)
{ return ((OBJECT*) a)->compare((OBJECT*) b); }

void OBJECT_LIST::sort()
{ list.sort(compare); }

// @@@ void OBJECT_LIST::load(PARSE* parse, coor pos)

void OBJECT_LIST::load(PARSE* parse, coor pos)
{
    for(;;)
    {
	OBJECT* object = ParseObject(parse, pos);
	
	if(object == NULL)
	    break;
	else
	    add(object);
    }
}

// @@@ void OBJECT_LIST::save(BINARY* binary)

void OBJECT_LIST::save(BINARY* binary)
{
    reset();

    while(remain() > 0)
    {
	current()->save(binary);
	next();
    }
    binary->PutShort(OBJECT_END); // Mark end of list (See OBJECT::save)
}

// @@@ void OBJECT_LIST::load(BINARY* binary)

void OBJECT_LIST::load(BINARY* binary)
{
    for(;;)
    {
	OBJECT* object = LoadObject(binary);
	
	if(object == NULL)
	    break;
	else
	    add(object);
    }
}

// @@@ OBJECT_LIST::OBJECT_LIST()

OBJECT_LIST::OBJECT_LIST()
{ }
// @@@ OBJECT_LIST::OBJECT_LIST(PARSE* parse, coor pos)

OBJECT_LIST::OBJECT_LIST(PARSE* parse, coor pos)
{ load(parse, pos); }

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
