// @ myth - server - view.C
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk

// @@ include

#include "server/view.h"
#include "common/error.h"
#include "server/player.h"
#include "server/global.h"

// @@ bool IsPermanent(char c)
//
// Return TRUE, if c should be placed permanently on the map.

inline bool IsPermanent(char c)
{ return !isalnum(c) && c != '@'; }

// @@ enum view_type

enum view_type
{
    VIEW_CLEAR = 0,
    VIEW_SEEN = 1,
    VIEW_BLOCKING = 2,
    VIEW_UNKNOWN = 4
};  

// @@ VIEW_LIST

// @@@ void VIEW_LIST::erase()

void VIEW_LIST::erase()
{
    delete VIEW_LIST::current();
    LIST::remove();
}

// @@@ void VIEW_LIST::clear()

void VIEW_LIST::clear()
{
    VIEW_LIST::reset();

    while(VIEW_LIST::check())
	VIEW_LIST::erase();
}

// @@ VIEW

// @@@ bool VIEW::CanSee(coor where)

bool VIEW::CanSee(coor where)
{
    if(home == NoWhere || ! (abs(where - home) < coor(sight, sight)))
	return FALSE;
    else
	return (view[where - SightCorner] & VIEW_SEEN) != 0;
}

// @@@ bool VIEW::CanDetect(coor where)

bool VIEW::CanDetect(coor where)
{
    if(home == NoWhere)
	return FALSE;
    if(detect == ERROR)
	return FALSE;
    if(sqr(where.x - home.x) + sqr(where.y - home.y) < DetectSquare)
	return TRUE;

    return FALSE;
}


// @@@ void VIEW::SetView(LEVEL* level, coor pos, bool light)

void VIEW::SetView(LEVEL* level, coor pos, bool light)
{
    // We already know this spot
    if(view[pos] != VIEW_UNKNOWN)
	return;

    bool blocking = FALSE;	// TRUE, iff something is blocking the view

    straight line(SightCenter, pos); // Check from SightCenter to pos
    line.next();		// (But not including SightCenter)

    while(line.check())
    {
	if(blocking)
	    // Something has blocked the view
	    view[line] = VIEW_CLEAR;
	else if(view[line] & VIEW_BLOCKING)
	    // This has been marked blocking before
	    blocking = TRUE;
	else if(view[pos] & VIEW_UNKNOWN)
	{
	    // Unknown, set it!
	    if(sqr(SightCenter.x - line.x) +
	       sqr(SightCenter.y - line.y) >= SightSquare)
	    {
		// This is out of sight
		view[line] = VIEW_CLEAR;	    
		blocking = TRUE;
	    }
	    else
	    {
		coor spot = SightCorner + line;
	    
		if(light || level->lit(spot))
		    view[line] |= VIEW_SEEN;
		else
		    view[line] = VIEW_CLEAR;

		if(IsRock(level->rock(spot)))
		{
		    view[line] |= VIEW_BLOCKING;
		    blocking = TRUE;
		}
		BUNCH* bunch = level->bunch(spot);

		if(bunch != NULL)
		{
		    if(bunch->check())
			bunch->update(level, spot, NULL);
		    if(bunch->empty())
			level->update(spot);
		    else
		    {
			if(bunch->light())
			    view[line] |= VIEW_SEEN;
			if(bunch->blocking())
			{
			    view[line] |= VIEW_BLOCKING;
			    blocking = TRUE;
			}
		    }
		}
	    }
	}
	line.next();
    }
}

// @@@ void VIEW::UpdateView(LEVEL* level, bool light)

void VIEW::UpdateView(LEVEL* level)
{
    if(home == NoWhere || sight < 1)
	return;

    const coor OutOfView = coor(sight * 2 - 1, sight * 2 - 1);
    region reg;			// Loop Index
    
    for(reg = region(Origo, OutOfView); reg.check(); reg.NextX())
	view[reg] = VIEW_UNKNOWN;

    BUNCH* bunch = level->create(home);
    bunch->update(level, home, NULL);
    bool light = bunch->light();

    for(reg = region(Origo, OutOfView); reg.check(); reg.NextX())
	SetView(level, reg, light);

    view[SightCenter] = VIEW_SEEN;
}


// @@@ void VIEW::send(coor where, char c, CHANNEL* channel) 

void VIEW::send(coor where, char c, CHANNEL* channel) 
{
    if(c == 0)
    {
	bug.error("VIEW::send", "zero is reserved");
	c = 127;
    }
    if(where != PlotStart + coor(index, 0) || index >= MessageSize)
    {
	flush(channel);
	PlotStart = where;
    }
    buffer[index++] = c;
}

// @@@ void VIEW::flush(CHANNEL* channel)

void VIEW::flush(CHANNEL* channel)
{
    switch(index)
    {
    case 0:
	/* Do Nothing */
	break;
    case 1:
	channel->PutChar(SERVER_PLOT);
	channel->PutCoor(PlotStart);
	channel->PutChar(buffer[0]);
	break;
    default:
	channel->PutChar(SERVER_PRINT);
	channel->PutCoor(PlotStart);
	for(int i = 0; i < index; i++)
	    channel->PutChar(buffer[i]);
	channel->PutChar(0);
    }
    index = 0;
}

// @@@ char VIEW::examine(LEVEL* level, coor where, bool SeeInvisible)

char VIEW::examine(LEVEL* level, coor where, bool SeeInvisible)
{
    if(CanSee(where))
	return level->symbol(where, SeeInvisible);

    BUNCH* bunch = level->bunch(where);

    if(bunch != NULL && CanDetect(where))
	return bunch->detect();

    return SYMBOL_ERROR;
}

// @@@ void VIEW::insert(coor where, char what, CHANNEL* channel)

void VIEW::insert(coor where, char what, CHANNEL* channel)
{
    if(IsPermanent(what))
	map[depth].plot(where, what);
    else
	list.insert(where, what);
    
    send(where, what, channel);
}

// @@@ void VIEW::update(LEVEL* level, PLAYER* player, CHANNEL* channel,

void VIEW::update(LEVEL* level, PLAYER* player, CHANNEL* channel,
		  bool SeeInvisible, mapping_type type, int length) 
{
    if(home == NoWhere)
	return;
    // How fare can you see?
    const int width = max(length, max(sight, detect));
    // Where on the map does we start
    const coor corner = home - coor(width - 1, width - 1);
    // Size of area where monsters can be detected
    const coor Size(width * 2, width * 2);

    // Find out which positions can be seen, has light etc.
    UpdateView(level);

    list.reset();

    for(region reg(Origo, Size); reg.check(); reg.NextX())
    {
	coor where = reg + corner;
	char what = examine(level, where, SeeInvisible);

	if(what == SYMBOL_ERROR && type != MAPPING_SIGHT &&
	   sqr(where.x - home.x) + sqr(where.y - home.y) <= sqr(length))
	{
	    BUNCH* bunch = level->bunch(where);
	    
	    switch(type)
	    {
	    case MAPPING_FORGET:
		if(!player->IsConfused())
		    what = ' ';
		break;
	    case MAPPING_OBJECT:
		if(bunch != NULL && !player->IsConfused())
		    what = bunch->look(SeeInvisible);
		break;
	    case MAPPING_ALL:
		if(!player->IsConfused())
		    what = level->symbol(where, SeeInvisible);
		break;
	    case MAPPING_SIGHT:
	    case MAPPING_ERROR:
	    default:
		bug.error("VIEW::mapping", "Impossible type");
	    }
	}
	while(list.check() &&
	      (list.where().y < where.y ||
	       (list.where().y == where.y && list.where().x < where.x)))
	{
	    send(list.where(), map[depth].what(list.where()), channel);
	    list.erase();
	}
	if(list.check() && list.where() == where)
	{
	    if(list.what() == what)
		list.next();
	    else
	    {
		if(what == SYMBOL_ERROR)
		    send(where, map[depth].what(where), channel);
		else
		    insert(where, what, channel);

		list.erase();
	    }
	}
	else if(what != SYMBOL_ERROR && map[depth].what(where) != what)
	    insert(where, what, channel);
    }
    while(list.check())
    {
	send(list.where(), map[depth].what(list.where()), channel);
	list.erase();
    }
    flush(channel);
}

// @@@ void VIEW::enter(level_type type, CHANNEL* channel)

void VIEW::enter(level_type type, CHANNEL* channel)
{
    // Remove all temporary ploted chars
    list.clear();
    // Select the new level
    depth = type;
    // Send the new map to the player
    channel->PutChar(SERVER_MAP);
    map[depth].save(channel);
}

// @@@ void VIEW::forget(CHANNEL* channel)

void VIEW::forget(CHANNEL* channel)
{
    map[depth].clear();
    enter(depth, channel);
}

// @@@ void VIEW::move(LEVEL*, coor pos, CHANNEL* channel)

void VIEW::move(LEVEL*, coor pos, CHANNEL* channel)
{
    if(home != pos)
    {
	home = pos;

	if(home == NoWhere)
	    return;

	SightCorner = home - SightCenter;
	DetectCorner = home - DetectCenter;
    }
    // Also used when restoring a player
    channel->PutChar(SERVER_HOME);
    channel->PutCoor(pos);
}

// @@@ void VIEW::plot(LEVEL* level, coor pos, bool SeeInvisible,

void VIEW::plot(LEVEL* level, coor pos, bool SeeInvisible,
		CHANNEL* channel = NULL)
{
    char what = level->symbol(pos, SeeInvisible);

    if(map[depth].what(pos) != what && IsPermanent(what))
    {
	map[depth].plot(pos, what);

	if(channel != NULL)
	{
	    channel->PutChar(SERVER_PLOT);
	    channel->PutCoor(pos);
	    channel->PutChar(what);
	}
    }
}

// @@@ void VIEW::look(int distance)

void VIEW::look(int distance)
{
    if(sight != distance)
    {
	sight = distance;
	SightSquare = sqr(sight);
	SightCenter = coor(sight - 1, sight - 1);
	SightCorner = home - SightCenter;
	if(sight > 0)
	    view.resize(coor(sight * 2 - 1, sight * 2 - 1));
    }
}

// @@@ void VIEW::detection(int distance)

void VIEW::detection(int distance)
{
    if(distance < 2)
	detect = ERROR;
    else if(detect != distance)
    {
	detect = distance;
	DetectSquare = sqr(detect);
	DetectCenter = coor(detect - 1, detect - 1);
	DetectCorner = home - DetectCenter;
    }
}

// @@@ void VIEW::save(BINARY* binary)

void VIEW::save(BINARY* binary)
{
    // Save each map
    for(level_type level = 0; level < global.DungeonSize(); level++)
	map[level].save(binary);

    // We don't save the list of temporary information, sorry

    // Save position
    binary->PutShort(depth);
    binary->PutCoor(home);

    // Save sight
    binary->PutShort(sight);

    // Save detection
    binary->PutShort(detect);
}

// @@@ void VIEW::load(BINARY* binary)

void VIEW::load(BINARY* binary)
{
    // Load each map
    for(level_type level = 0; level < global.DungeonSize(); level++)
	map[level].load(binary);

    // Clear the list of temporary information
    list.clear();

    // Load position
    depth = binary->GetShort();
    home = binary->GetCoor();

    // Load sight
    look(binary->GetShort());

    // Load detection
    detection(binary->GetShort());

}

// @@@ VIEW::VIEW()

VIEW::VIEW()
{
    depth = LEVEL_ERROR;
    home = NoWhere;
    sight = ERROR;
    look(1);
    detect = ERROR;
    detection(0);
    index = 0;
    PlotStart = NoWhere;
    map = new MAP[global.DungeonSize()];
    
    // This should not be necessary
    for(level_type type = 0; type < global.DungeonSize(); type++)
	map[type].SelectDefault(' ');
}

// @@@ VIEW::~VIEW()

VIEW::~VIEW()
{
    delete map;
}

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
